/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/


#include <WCore>
#include "PFractal.h"
//#include "Fractal.h"

#if 0//GENERATING68881
	#define USE68ASM
#endif

typedef struct
{
	bool *pfTerm;
	ulong escLim;
	AComplex pc;
	double cx;
	double cy;
} PFractalMandelbrotJuliaData;

#ifdef USE68ASM
	#include "A68Mandelbrot.h"
	#define GetMandel AsmGetMandelbrot
#endif

#ifdef _X86_
	#define USE86ASM
#endif

#ifdef USE86ASM
static void AsmGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim);
#endif

static void CGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim);
static void CGetJuliaMandel(AComplex param, AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim, bool *pfTerm);

#ifndef USE68ASM
static ulong GetMandel(ulong, AComplex);


static ulong GetMandel(ulong escLim, AComplex pc)
{
	AComplex x;
	AScalar r;
	ulong i;

	r = 0;
	ACMake(x,0,0);
	for (i=escLim; (i>0) && (r <4.0); i--)
	//for (i=0; (i<escLim) && (r <4.0); i++)
	{                     
		// x = x^2 +c;
		ACSquare(x);
		ACAdd(pc,x);
		r = ACMagnitudeSqr(x);             

		ThreadYieldAfterPeriod(AGetThreadTimeoutFrequency());
	}        
	
	return i;                                                         	
	//return (escLim - i);
}

#endif

static ulong GetJuliaMandel(PFractalMandelbrotJuliaData*);
static ulong GetJuliaMandel(PFractalMandelbrotJuliaData *pfmjd)
{
	AComplex x;
	AScalar r;
	ulong i;

	r = 0;
	ACMake(x,pfmjd->cx, pfmjd->cy);
	for (i=pfmjd->escLim; i && !(*pfmjd->pfTerm) && (r <4.0); i--)
	{                     

		// x = x^2 +c;
		ACSquare(x);
		ACAdd(pfmjd->pc,x);
		r = ACMagnitudeSqr(x);             


		ThreadYieldAfterPeriod(AGetThreadTimeoutFrequency());
	}                                                                     
	
	return i;
}

static void CGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * buff ,ulong xLim)
{
	AComplex c;

	ACMake(c, *pcR, *pcI);

	switch(pixdepth)
	{
		case 1:
		{
			uchar *b;
		
			b = buff;
			while(xLim--)
			{
				*b++ = (uchar) GetMandel(escLim, c);
				ACReal(c) += *deltax;			
			}
		}
		break;

		case 2:
		{
			ushort *b;
		
			b = (ushort*)buff;
			while(xLim--)
			{
				*b++ = (ushort) GetMandel(escLim, c);
				ACReal(c) += *deltax;
			}
		}
		break;

		case 3:
		{
			ARGB24 *b;
		
			b = (ARGB24*)buff;
			while(xLim--)
			{
				ulong val;

				val = GetMandel(escLim, c);
				*b++ = *(ARGB24*)&val;
				ACReal(c) += *deltax;
			}
		}
		break;

		case 4:
		{
			ulong *b;
			
			b = (ulong*)buff;
			while(xLim--)
			{
				*b++ = (ulong) GetMandel(escLim, c);
				ACReal(c) += *deltax;
			}
		}
		break;
	}
}

static void CGetJuliaMandel(AComplex param, AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * buff ,ulong xLim, bool *pfTerm)
{
	PFractalMandelbrotJuliaData fmjd;

	ACCopy(param, fmjd.pc);
	fmjd.cx = *pcR;
	fmjd.cy = *pcI;
	fmjd.pfTerm = pfTerm;
	fmjd.escLim = escLim;
	switch(pixdepth)
	{
		case 1:
		{
			uchar *b;
	
			b = buff;
			while(xLim--)
			{
				*b++ = (uchar) GetJuliaMandel(&fmjd);
				fmjd.cx += *deltax;			
			}
		}
		break;

		case 2:
		{
			ushort *b;
	
			b = (ushort*)buff;
			fmjd.escLim = 0xffff;
			while(xLim--)
			{
				*b++ = (ushort) GetJuliaMandel(&fmjd);
				fmjd.cx += *deltax;			
			}
		}
		break;

		case 3:
		{
			ARGB24 *b;
		
			b = (ARGB24*)buff;
			while(xLim--)
			{
				ulong val;

				val = GetJuliaMandel(&fmjd);
				*b++ = *(ARGB24*)&val;
				fmjd.cx += *deltax;			
			}
		}
		break;

		case 4:
		{
			ulong *b;
		
			b = (ulong*)buff;
			while(xLim--)
			{
				*b++ = (ulong) GetJuliaMandel(&fmjd);
				fmjd.cx += *deltax;			
			}
		}
		break;
	}
}

void PFGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim)
{
	Assert(pcR);
	Assert(pcI);
	Assert(deltax);
	Assert(escLim);
	Assert(dOffset);
	Assert(xLim);
	
#ifdef USE86ASM
	AsmGetMandel(pcR, pcI, deltax, pixdepth, escLim, dOffset, xLim);
#else
	CGetMandel(pcR, pcI, deltax, pixdepth, escLim, dOffset, xLim);
#endif

}

void PFGetJuliaMandel(AComplex param, AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim, bool *pfTerm)
{
	Assert(pcR);
	Assert(pcI);
	Assert(deltax);
	Assert(pixdepth);
	Assert(dOffset);
	Assert(xLim);
	
//#ifdef USE86ASM
//	AsmGetMandel(pcR, pcI, deltax, escLim, dOffset, xLim);
//#else
	CGetJuliaMandel(param, pcR, pcI, deltax, pixdepth, escLim, dOffset, xLim, pfTerm);
//#endif

}

void PFSubtractArray(uchar * pBuffer, ulong count, long increment, short numBytes)
{
	Assert(pBuffer);
	Assert(count);
	Assert(numBytes);
	
	if (increment == 0)
		return;
		
	switch(numBytes)
	{
		case 1:
			while (count--)
				((uchar*)pBuffer)[count] = increment - ((uchar*)pBuffer)[count];
		break;
		case 2:
			while (count--)
				((ushort*)pBuffer)[count] = increment - ((ushort*)pBuffer)[count];
		break;
		case 3:
			while (count--)
				*(ulong *)&((ARGB24*)pBuffer)[count] = increment - *(ulong *)&((ARGB24*)pBuffer)[count];
		break;
		case 4:
			while (count--)
				((ulong*)pBuffer)[count] = increment - ((ulong*)pBuffer)[count];
		break;
		
		default:
			Assert(false);
		break;
	}
}


#ifndef USE86ASM
void PFAddArray(uchar * pBuffer, ulong count, long increment, short numBytes)
{
	Assert(pBuffer);
	Assert(count);
	Assert(numBytes);
	
	if (increment == 0)
		return;
		
	switch(numBytes)
	{
		case 1:
			while (count--)
				((uchar*)pBuffer)[count] += increment;
		break;
		case 2:
			while (count--)
				((ushort*)pBuffer)[count] += increment;
		break;
		case 3:
			while (count--)
				*(ulong *)&((ARGB24*)pBuffer)[count] += increment;
		break;
		case 4:
			while (count--)
				((ulong*)pBuffer)[count] += increment;
		break;
		
		default:
			Assert(false);
		break;
	}
}


#else
void PFAddArray(uchar * pBuffer, ulong count, long increment, short numBytes)
{
_asm
{
	push	eax
	push	ebx
	push	ecx
	push	edx

	mov		ecx, count		//loop counter
	mov		eax, increment
	mov		ebx, pBuffer	//Buffer

	cmp		numBytes, 1		//One byte buffer?
	je		onebyteloop		//If not execute the code for two byte buffer
	cmp		numBytes, 2		//Two byte buffer?
	je		twobyteloop
	cmp		numBytes, 3		//Three byte buffer
	je		threebyteloop
	jmp		fourbyteloop

	even
	onebyteloop:
	add		byte ptr [ebx], al	//add the number
	inc		ebx
	loop	onebyteloop
	jmp		doneaddarray

	even
	twobyteloop:
	add		word ptr [ebx], ax	//add the number
	add		ebx, 2
	loop	twobyteloop
	jmp		doneaddarray

	even
	threebyteloop:
	mov		edx, dword ptr[ebx]
	add		edx, eax
	mov		word ptr[ebx], dx
	shr		edx, 8
	mov		byte ptr[ebx+2], dh
	add		ebx, 3
	
	loop	threebyteloop
	jmp		doneaddarray

	even
	fourbyteloop:
	add		dword ptr [ebx], eax	//add the number
	add		ebx, 4
	loop	fourbyteloop
	jmp		doneaddarray

	doneaddarray:

	pop		edx
	pop		ecx
	pop		ebx
	pop		eax
}
}


static void AsmGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim)
{
	long double env[20];
_asm
{
	push	eax
	push	ebx
	push	ecx
	push    edx

	fsave   byte ptr env
	
	xor		edx,edx
	mov     dx, pixdepth      ; save in dx for later use

	fldz                    ; temp		5
	mov     ebx,pcR
	fld     qword ptr [ebx] ;c.real		4
	mov     ebx,pcI
	fld     qword ptr [ebx] ;c.img	3
	fld1    
	fadd    st, st          
	fadd    st, st          ; 4     2
	fldz                    ; x.real        1
	fldz                    ; x.img         0

	xor		ecx,ecx
	mov     ecx,xLim
	mov     ebx, dOffset    ; ebx now contains the current pixel buffer ptr
xloop:
	push    ecx
	mov     ecx, escLim
even
mloop:
	; ACSquare(x) Incorrect
	; First tackle the imaginary part
	fst     st(5)           ; temp = x.img
	fadd    st,st
	fmul    st,st(1)        ; x.img *= x.real
	; Second the real part
	fxch    st(1)           ; exch(x.img, xreal)
	fxch    st(5)           ; exch(x.real, temp)
	fmul    st,st           ; temp ^= 2
	fxch    st(5)           ; exch(x.real, temp)
	fmul    st,st           ; x.real ^=2
	fsub    st,st(5)       ; x.real -= temp

	; ACAdd(*pc, x)
	fadd    st, st(4)       ; x.real += pc->real
	fxch    st(1)           ; exch(x.real, x.img
	fadd    st, st(3)       ; x.img += pc->img

	; ACMagnitude(x)
	fld     st              ; push r = x.img
	fmul    st,st           ; r ^= 2
	fstp    st(6)           ; temp = r
	fld     st(1)           ; push r = x.img
	fmul    st,st           ; r ^= 2
	fadd    st,st(6)       ; r += temp

	; Compare r to 4.0
	fcomp   st(3)           ; st(3)==4.0
	fstsw   ax
	sahf
	jnc     endloop

	loop mloop

endloop:
	; Result is in CX
	; Are we stuffing in a 1-byte or 2-byte buffer?
	;  depending on the size of escLim, we increment by proper byte delta
	cmp		dx, 1		; Byte sized chunks?
	jg		notbytesized
	mov		byte ptr[ebx], cl ; Only the lower byte has valid data
	jmp		finishwritingpixel

notbytesized:
	cmp		dx, 2		; 16-bit pixel?
	jg		pixelis24bit
	mov     word ptr [ebx], cx
	jmp		finishwritingpixel

pixelis24bit:	// doesn't matter if its 24-bit or 32 bit
	mov     dword ptr [ebx], ecx

finishwritingpixel:
	add		ebx, edx

	; increment x by deltax
	mov     eax,deltax
	fld     qword ptr [eax]
	faddp   st(5),st

	fsub    st,st
	fst     st(1)
	pop     ecx
	loop    xloop

	frstor  byte ptr env[0]

	pop     edx
	pop		ecx
	pop		ebx
	pop		eax
}
}

#if 0
static void AsmGetMandel(AScalar *pcR, AScalar *pcI, AScalar *deltax, ushort pixdepth, ulong escLim, uchar * dOffset ,ulong xLim)
{
	long double env[20];
_asm
{
	push	eax
	push	ebx
	push	ecx
	push    edx

	fsave   byte ptr env
	
	xor		edx,edx
	mov     dx, pixdepth      ; save in dx for later use

	fldz                    ; temp		5
	mov     ebx,pcR
	fld     qword ptr [ebx] ;c.real		4
	mov     ebx,pcI
	fld     qword ptr [ebx] ;c.img	3
	fld1    
	fadd    st, st          
	fadd    st, st          ; 4     2
	fldz                    ; x.real        1
	fldz                    ; x.img         0

	xor		ecx,ecx
	mov     ecx,xLim
	mov     ebx, dOffset    ; ebx now contains the current pixel buffer ptr
xloop:
	push    ecx
	mov     ecx, escLim
even
mloop:
	; ACSquare(x) Incorrect
	; First tackle the imaginary part
	fst     st(5)           ; temp = x.img
	fadd    st,st
	fmul    st,st(1)        ; x.img *= x.real
	; Second the real part
	fxch    st(1)           ; exch(x.img, xreal)
	fxch    st(5)           ; exch(x.real, temp)
	fmul    st,st           ; temp ^= 2
	fxch    st(5)           ; exch(x.real, temp)
	fmul    st,st           ; x.real ^=2
	fsub    st,st(5)       ; x.real -= temp

	; ACAdd(*pc, x)
	fadd    st, st(4)       ; x.real += pc->real
	fxch    st(1)           ; exch(x.real, x.img
	fadd    st, st(3)       ; x.img += pc->img

	; ACMagnitude(x)
	fld     st              ; push r = x.img
	fmul    st,st           ; r ^= 2
	fstp    st(6)           ; temp = r
	fld     st(1)           ; push r = x.img
	fmul    st,st           ; r ^= 2
	fadd    st,st(6)       ; r += temp

	; Compare r to 4.0
	fcomp   st(3)           ; st(3)==4.0
	fstsw   ax
	sahf
	jnc     endloop

	loop mloop

endloop:
	; Result is in CX
	; Are we stuffing in a 1-byte or 2-byte buffer?
	;  depending on the size of escLim, we increment by proper byte delta
	cmp		dx, 1		; Byte sized chunks?
	jg		notbytesized
	mov		byte ptr[ebx], cl ; Only the lower byte has valid data
	jmp		finishwritingpixel

notbytesized:
	cmp		dx, 2		; 16-bit pixel?
	jg		pixelis24bit
	mov     word ptr [ebx], cx
	jmp		finishwritingpixel

pixelis24bit:	// doesn't matter if its 24-bit or 32 bit
	mov     dword ptr [ebx], ecx

finishwritingpixel:
	add		ebx, edx

	; increment x by deltax
	mov     eax,deltax
	fld     qword ptr [eax]
	faddp   st(5),st

	fsub    st,st
	fst     st(1)
	pop     ecx
	loop    xloop

	frstor  byte ptr env[0]

	pop     edx
	pop		ecx
	pop		ebx
	pop		eax
}
}
#endif

#endif
